#!/usr/local/bin/perl
##############################################################################
# Author: Binny V Abraham                                                    #
# Website: http://www.geocities.com/binnyva                                  #
# E-Mail : binnyva@hotmail.com                                               #
# Get more Perl scripts from http://www.geocities.com/binnyva/code/perl      #
##############################################################################
# Name    : SedGUI                                                           #
# Version : 1.00.A Beta                                                      #
# Date    : December 10, 2005 - December 26, 2005                            #
# This is a GUI version of the Sed(Stream EDitor) program in Unix. Sed is a  #
# very powerful and very useful program - that sed, it is very user          #
# unfriendly. So this is a GUI version for the Sed progarm.                  #
#                                                                            #
# Get more Perl scripts from http://www.geocities.com/binnyva/code/perl      #
##############################################################################
use strict;

############################# Global Variables ###############################
#Program Information
my %program = (
	name	=>	"SedGUI",
	author	=>	"Binny V Abraham",
	version	=>	"1.00.A Beta",
	email	=>	"binnyva\@hotmail.com",
	website	=>	"http://www.geocites.com/binnyva"
);
my %opt = (
	start_folder		=> "~",
	start_script_folder => "~",
	option_file			=> "~/.sedgui",
	print_all_lines		=> 0,	#If 1, Prints all the lines - not just the ones that matched the regexp
	print_unmatched		=> 0,	#If 1, Prints the lines the DONT match the regexp
	show_progress		=> 1,
	show_line_numbers	=> 1,
	word_wrap			=> 0,
	editor				=> 'vi',
	direct_command		=> 2	#If 1, 'direct_command' will always be on. If 0, it will always be off. 
);								#	If anyother value, the sytem will try to find wether the regexp 
								#	is a direct command and act accodingly.
								#	If the direct_command mode is enabled, just run the command as they come 
								#	from the regexp field. This can be in the form of 's/<a href="(.+?)"/\1/gi;'
								#	The system will do the replace with just one field.

my $file;
my $script_file;
my $regexp = "";
my $replace = "";
my $modifiers = '';
my $filter = "";

#Others
my $percent_done = 0;
my $top_history_list;	#A top level window that appears when the history of the searches are called up.
my $slst_history_list;

my @regexp_history= ();	#The list holding all the previous regexps we have entered
my $regexp_history_selected = 0;

my @replace_history= ();
my $replace_history_selected = 0;

my $gui_is_on = 1;

######################### Command Line Arguments #############################
#Get the needed command line arguments
if(@ARGV) {
	$file = $ARGV[0]; #The first argument is the file.

	#A script file can be specified as the second argument - if so, execute it and exit.
	if($ARGV[1]) {
		$gui_is_on = 0;
		$script_file = $ARGV[1];
		my $data = openScriptFile();#This function will call the sed function that does the job
		print $data;
		exit;
	}
}

############################### GUI Building #################################
use Tk;

#More Tk stuff
use Tk::ProgressBar;
use Tk::DropSite;
use Tk::TextUndo;

if($^O eq 'MSWin32') {
	#Get Position of options file
	use File::Basename;
	$opt{'option_file'} = dirname($0) . "/.sedgui";
	$opt{'editor'} = "notepad";
}

# Main Window
my $mw = new MainWindow;

my $lab_title = $mw -> Label(text=>"$program{'name'} V$program{'version'}", -font=>"ansi 14 bold");

my $lab_file = $mw -> Label(text=>"File");
my $ent_file = $mw -> Entry(textvariable=>\$file,-width=>65);
my $but_file = $mw -> Button(text=>'Browse...',command=>\&getFile);

my $lab_regexp = $mw -> Label(text=>"RegExp");
my $ent_regexp = $mw -> Entry(textvariable=>\$regexp);
my $but_escape = $mw -> Button(text=>"Escape",-command=>\&escape);

my $lab_replace= $mw -> Label(text=>"Replace With");
my $ent_replace= $mw -> Entry(textvariable=>\$replace);
my $lab_mods   = $mw -> Label(text=>"Modifiers");
my $ent_mods   = $mw -> Entry(textvariable=>\$modifiers);

my $lab_filter = $mw -> Label(text=>"Filter");
my $ent_filter = $mw -> Entry(textvariable=>\$filter);
my $but_filter = $mw -> Button(-text=>"Filter",-command=>\&filterData);

my $frm_text = $mw ->Frame;
my $lst_line_numbers = $frm_text->Listbox(-width=>4,-bg=>'#808080',-border=>0);

# The main text field
my $stxt_text = $frm_text->Scrolled('TextUndo',-scrollbars=>'se',-spacing2=>2,-spacing3=>2);
$stxt_text->configure(-wrap=>'none') unless $opt{'word_wrap'};
#Geometry
$lst_line_numbers -> grid(qw/-column 0 -row 0 -sticky ns/);
$stxt_text -> grid(qw/-column 1 -row 0 -sticky nesw/);

my $slb_y_text = $stxt_text->Subwidget('yscrollbar');
my $widgets_to_scroll = [$lst_line_numbers,$stxt_text];
# Configure each Listbox to call &scroll_listboxes
foreach my $widget (@$widgets_to_scroll) {
	$widget->configure(-yscrollcommand => [ \&scrollMultiple, $slb_y_text, $widget, $widgets_to_scroll ]);
}

# Configure the Scrollbar to scroll each Listbox
$slb_y_text -> configure(-command => sub { foreach my $widget (@$widgets_to_scroll) {
          $widget -> yview(@_);
        }});

my $but_process = $mw -> Button(text=>"Process",-command=>\&sed);

my $pbr_progress = $mw -> ProgressBar(-borderwidth => 2, -anchor => 'w', -gap => 2,
    -from => 0, -to => 100, -blocks => 20, -colors => [0,'darkblue']
);


### Geometry Management
$lab_title -> grid(qw/-row 0 -column 0 -columnspan 3/);

$lab_file -> grid(qw/-row 1 -column 0 -sticky w/);
$ent_file -> grid(qw/-row 1 -column 1 -sticky we/);
$but_file -> grid(qw/-row 1 -column 2 -sticky we/);

$lab_regexp -> grid(qw/-row 2 -column 0 -sticky w/);
$ent_regexp -> grid(qw/-row 2 -column 1 -sticky we/);
$but_escape -> grid(qw/-row 2 -column 2 -sticky we/);

$lab_replace-> grid(qw/-row 3 -column 0 -sticky w/);
$ent_replace-> grid(qw/-row 3 -column 1 -columnspan 2 -sticky we/);
$lab_mods   -> grid(qw/-row 4 -column 0 -sticky w/);
$ent_mods   -> grid(qw/-row 4 -column 1 -columnspan 2 -sticky we/);

$lab_filter -> grid(qw/-row 5 -column 0 -sticky w/);
$ent_filter -> grid(qw/-row 5 -column 1 -sticky we/);
$but_filter -> grid(qw/-row 5 -column 2 -sticky we/);

$but_process -> grid(qw/-row 6 -column 0 -columnspan 3/);

$frm_text -> grid(qw/-row 7 -column 0 -columnspan 3 -sticky nswe/);
$pbr_progress -> grid(qw/-row 8 -column 0 -columnspan 3 -sticky we/);

#######################  Keyboad shortcuts... ########################
$mw -> bind('<Key-Escape>',   sub { exiter(); });
$mw -> bind('<Control-Key-w>',sub { exiter(); });
$mw -> bind('<Key-F1>', sub { about(); });
$mw -> bind('<Control-Key-o>', sub { getFile(); });
$mw -> bind('<Control-Key-s>', sub { saveFile(); });
$mw -> bind('<Control-Key-e>', sub { editor($file); });
$mw -> bind('<Control-Key-O>', sub { getScriptFile(); });
$mw -> bind('<Control-Key-S>', sub { saveScriptFile(); });
$mw -> bind('<Control-Key-E>', sub { editor($script_file); });
$mw -> bind('<Key-F5>', sub { openFile(); });

#Other Bindings
$ent_regexp -> bind('<Up>', 	sub { histSelect(-1,$ent_regexp);});
$ent_regexp -> bind('<Down>',	sub { histSelect(1,$ent_regexp); });
$ent_regexp -> bind('<Return>',	sub { showHistory($ent_regexp);  });

$ent_replace -> bind('<Up>', 	sub { histSelect(-1,$ent_replace);});
$ent_replace -> bind('<Down>',	sub { histSelect(1,$ent_replace); });
$ent_replace -> bind('<Return>',sub { showHistory($ent_replace);  });

$mw -> resizable(0,0);
$mw -> wm('protocol','WM_DELETE_WINDOW', sub { &exiter; });
$mw -> wm('title',"$program{'name'} v$program{'version'}");

#Call the 'main' like function
init();

#Catch Drops
# The allowed drag and drop types on Unix systems are "KDE", "XDND" and "SUN"
# and on Windows systems the "Win32" dnd type.
# When dropping occurs, execute the accept_drop callback.
$mw->DropSite(-dropcommand => [\&accept_drop, $mw],
	-droptypes => ($^O eq 'MSWin32' ? 'Win32' : ['KDE', 'XDND', 'Sun'])
);

##### Menu
my $mbar = $mw -> Menu();
my $menu_font = "ansi";

#The Main Buttons
my $mnu_file= $mbar -> cascade(-label =>"File",   -underline=>0, -tearoff => 0);
my $mnu_opt = $mbar -> cascade(-label =>"Options",-underline=>0, -tearoff => 0);
my $mnu_help= $mbar -> cascade(-label =>"Help",   -underline=>0, -tearoff => 0);

#File Menu
$mnu_file -> command(-label => "~Open File", -font=>$menu_font, -command=>\&getFile, -accelerator=>"Ctrl+O");
$mnu_file -> command(-label => "~Save File", -font=>$menu_font, -command=>\&saveFile, -accelerator=>"Ctrl+S");
$mnu_file -> command(-label => "Re~fresh", -font=>$menu_font, -command=>\&openFile, -accelerator=>"F5");
$mnu_file -> separator;
$mnu_file -> command(-label => "Open Script", -font=>$menu_font, -command=>\&getScriptFile, -accelerator=>"Ctrl+Shift+O");
$mnu_file -> command(-label => "Save Script", -font=>$menu_font, -command=>\&saveScriptFile, -accelerator=>"Ctrl+Shift+S");
$mnu_file -> separator;
$mnu_file -> command(-label => "E~xit",-command=> sub { exit; }, -font=>$menu_font, -accelerator=>"Esc");

#Options
$mnu_opt -> checkbutton(-label => "Show Only Matched Lines", -font=>$menu_font,
	-variable=>\$opt{'print_all_lines'},	-command=>sub{ $opt{'print_unmatched'}=0; } );
$mnu_opt -> checkbutton(-label => "Show Unmatched Lines", -font=>$menu_font, 
	-variable=>\$opt{'print_unmatched'},-command=>sub{ $opt{'print_all_lines'}=0; } );
$mnu_opt -> separator;
$mnu_opt -> checkbutton(-label => "Word Wrap", -font=>$menu_font, -variable=>\$opt{'word_wrap'},
	-command => sub {
		if($opt{'word_wrap'}) {
			$stxt_text->configure(-wrap=>'char');
		} else {
			$stxt_text->configure(-wrap=>'none');
		}
		lineNumbersDisplay();
	}
);
$mnu_opt -> checkbutton(-label => "Show Progress", -font=>$menu_font, -variable=>\$opt{'show_progress'});
$mnu_opt -> checkbutton(-label => "Show Line Numbers", -font=>$menu_font, -variable=>\$opt{'show_line_numbers'},
	-command=>\&lineNumbersDisplay);
#$mnu_opt -> checkbutton(-label=>"",-font=>$menu_font,-variable=>\$opt{''});

#Help
$mnu_help -> command(-label => "~About", -font=>$menu_font, -command=>\&about, -accelerator=>"F1");
$mw ->configure(-menu => $mbar);#Set this as menu

MainLoop;

################################# Functions ####################################

############################## Common Functions ################################
#Shows a message box with the specified message.
sub alert {
	my ($message) = @_;
	if($gui_is_on) {
		$mw->messageBox(-message=>$message);
	} else {
		print $message;
	}
}

#Shows a message box with the specified ERROR message.
sub error {
	my ($message) = @_;
	
	if($gui_is_on) {
		$mw->messageBox(-message=>$message,-icon=>'error');
	} else {
		print $message;
	}
}

#Shows a confirmation box with yes/no buttons
sub confirm {
	my ($message) = @_;
	my $response;
	if($gui_is_on) {
	 	$response = $mw->messageBox(-message=>$message,-type=>"YesNo",-icon=>"question");
 	} else {
		#print $message . "\nAssuming 'Yes'.\n";
 		#$response = <STDIN>;
 		#chomp($response);
 		#lc($response);
 		$response = "yes";
	}
 	if($response eq "yes" or $response eq "y") {
	 	return 1;
 	}
	return 0;
}

#Show a About box
sub about {
	$mw -> messageBox(-message =>"$program{'name'} V $program{'version'}
by Binny V Abraham

$program{'email'}
$program{'website'}",-title=>"About $program{'name'} V$program{'version'}");
}

############################## File Operations ##########################################
#Get the HTML file
sub getFile {
	my ($widget) = $ent_file;
	my $types = [
    	['HTML Files', ['.htm', '.html', '.shtml', '.php']],
    	['All Files',  '*']
	];
	
	my $folder = $opt{'start_folder'};
	my $last_file = $file;
	$folder =~ s/\//\\/g; #Convert all '/' to '\\'. Otherwise the saveFile function won't take it.
	$last_file =~ s/\//\\/g;
	
	my $temp_file = $mw -> getOpenFile(-title=>"Choose a File...",-filetypes=>$types,-initialfile=>$last_file,-initialdir=>$folder);
	if ($temp_file ne "") {
		$widget -> delete(0,'end');
		$widget -> insert(0,$file);
		$file = $temp_file;

		openFile();
	}
}

#Open the script file. A script file is a simple text file with a series of regexps and the replaces.
#	A regexps/replaces is given per line. First there will be the regexps - then a label 'Replaces' 
#	and then the replaces will occur
sub getScriptFile {
	my $types = [
    	['Sed Files', ['.sed', '.txt']],
    	['All Files',  '*']
	];
	
	my $folder = $opt{'start_script_folder'};
	my $last_file = $script_file;
	$folder =~ s/\//\\/g; #Convert all '/' to '\\'. Otherwise the saveFile function won't take it.
	$last_file =~ s/\//\\/g;
	
	my $temp_file = $mw -> getOpenFile(-title=>"Choose a File...",-filetypes=>$types,-initialfile=>$last_file,-initialdir=>$folder);
	if ($temp_file ne "") {
		$script_file = $temp_file;
		openScriptFile();
	}
}

#Open the script file and ask the user whether he/she wishes to execute the contents.
sub openScriptFile {
	#Empty the lists
	@regexp_history  = ();
	$regexp_history_selected = 0;
	@replace_history = ();
	$replace_history_selected = 0;
	
	#Open the files
	if(open(SCRIPT,$script_file)) {
		my ($reg,$rep);
		my $replace_found = 0;
		while(<SCRIPT>) {
			#Find the labels
			next if /^\-{20}$/;
			next if /^Regular Expressions$/;
			if (/^Replaces$/) {
				$replace_found++;
				next;
			}

			chomp;
			next unless($_);#Skip empty lines.

			#Add the regexps and replaces to history list
			if($replace_found) {
				push(@replace_history,$_);
			} else {
				push(@regexp_history,$_);
			}
		}
		#Run the script if only regexps where found.
		unless($replace_found) {
			$replace = "";
			if(confirm("Do you want to run this script on the current file?")) {
				return sed(@regexp_history);
			}
		}
	} else {
		error("Cannot open file '$script_file' : $!");
	}
}
# Save the history of all replaces/regexps to a script file. A script file is a simple text file 
#	with a series of regexps and the replaces.
#	A regexps/replaces is given per line. First there will be the regexps - then a label 'Replaces' 
#	and then the replaces will occur
sub saveScriptFile {
	my $types = [
    	['Sed Files', ['.sed', '.txt']],
    	['All Files',  '*']
	];

	my $folder = $opt{'start_script_folder'};
	my $last_file = $script_file;
	$folder =~ s/\//\\/g; #Convert all '/' to '\\'. Otherwise the saveFile function won't take it.
	$last_file =~ s/\//\\/g;

	my $temp_file = $mw -> getSaveFile(-title=>"Choose a File...",-filetypes=>$types,
		-initialfile=>$last_file, -initialdir=>$folder, -defaultextension=>'sed');
	if ($temp_file ne "") {
		$script_file = $temp_file;

		if(open(SCRIPT,">$script_file")) {
			my $replace_count = scalar(@replace_history);
			print SCRIPT "Regular Expressions\n--------------------\n" if ($replace_count);
			print SCRIPT join("\n",@regexp_history);
			if($replace_count) {
				print SCRIPT "\n\nReplaces\n--------------------\n";
				print SCRIPT join("\n",@replace_history);
			}
		} else {
			error("Cannot save to file '$script_file' : $!");
		}
		close(SCRIPT);
	}
}

#Read the file
sub openFile {
	my @lines;
	if(!open (FILE,$file)) {
		error("Can't open '$file': $!");
	} else {
		@lines=<FILE>;
		close(FILE);

		if($gui_is_on) {
			$stxt_text -> delete('1.0','end');
			$stxt_text -> insert('1.0',join(/\n/,@lines));

			lineNumbersDisplay();
		}
	}
	return @lines;
}
#Write the changed data back to the file
sub saveFile {
	unless(open (OUT, ">$file")) {
		error("Can't open '$file' for writing: $!");
	} else {
		print OUT getData();
	}
	close(OUT);
}

############################## GUI Functions ##########################################
#Scroll multiple widgets with just one scrollbar.
sub scrollMultiple {
	my ($sb, $scrolled, $lbs, @args) = @_;
	$sb->set(@args); # tell the Scrollbar what to display
	my ($top, $bottom) = $scrolled->yview();
	foreach my $list (@$lbs) {
		$list -> yviewMoveto($top); # adjust each lb
	}
}
#Get the name of the file that is dropped into the app.
sub accept_drop {
    my($widget, $selection) = @_;

    my $filename;
    eval {
	if ($^O eq 'MSWin32') {
	    $filename = $widget->SelectionGet(-selection => $selection,'STRING');
	} else {
	    $filename = $widget->SelectionGet(-selection => $selection,'FILE_NAME');
	}
    };
    if (defined $filename) {
		$file = $filename;
		openFile();
    }
}

#Selected the previously typed regexps by using the UP and Down arrow keys
sub histSelect {
	my ($direction,$widget) = @_;
	my $currnt_data = $widget -> get();
	
	my (@history,$selected_history);
	#Both the ent_regexp and ent_replace widgets go thru this function
	#	so I have to write the code for both.
	if($widget == $ent_regexp) {
		@history = @regexp_history;
		$selected_history = $regexp_history_selected;
	} else {
		@history = @replace_history;
		$selected_history = $replace_history_selected;
	}
	my $selected = $history[$selected_history];
		
	#Increment or decrement the selected item - based on the direction.
	if($selected eq $currnt_data) {
		if($direction == 1) {
			$selected_history++ unless($selected_history == $#history);
		} else {
			$selected_history-- unless($selected_history == 0);
		}
	} else {
		if($direction == 1) {
			$selected_history = 0;
		} else {
			$selected_history = $#history;
		}
	}
	
	#Now save the changed information back to the global variables.
	if($widget == $ent_regexp) {
		$regexp_history_selected = $selected_history;
	} else {
		$replace_history_selected = $selected_history;
	}

	$selected = $history[$selected_history];
	$widget -> delete(0,'end');
	$widget -> insert(0,$selected);
}

#The following functions will show the History list for the typed regular expressions...
# showHistory,histPick,histKill,histDestroy,centerList
#Copied from PTKSh V2.0
sub showHistory {
	#Tk->break if defined($top_history_list);
	my $widget = shift;
	
	$top_history_list = $mw->Toplevel;
	$top_history_list->overrideredirect(1); #Cool effect - but risky
	$top_history_list->title('History');
	$slst_history_list = $top_history_list->ScrlListbox(-scrollbars => 'oe',
              -width => 30, -height => 10,-bg=>'white')->pack;
	centerList($top_history_list,$widget);
	$slst_history_list->insert('end', @regexp_history);
	$slst_history_list->see(0);
	$slst_history_list->activate(0);
	$top_history_list->bind('<Double-1>',sub { histPick($widget); });
	$top_history_list->bind('<Return>',  sub { histPick($widget); });
	$top_history_list->bind('<Escape>',  sub { histKill(); });
	$top_history_list->bind('<Map>',	 sub { centerList($top_history_list,$widget);} );
	$top_history_list->bind('<Destroy>', sub { histDestroy($widget); });
	$top_history_list->focus;
	$slst_history_list->focus;
	$top_history_list->grab;
	Tk->break;
}

#If the user selects any command to be the active regular expression
sub histPick {
	my $item = $slst_history_list->get('active');
	return unless $item;

	my $widget = shift;

	$widget -> delete(0,'end');
	$widget -> insert('end',$item);
	$mw -> focus;
	$widget -> focus;
	histKill();
}

#Kill the History list.
sub histKill {
	if ($top_history_list) {
		$top_history_list->grabRelease;
		$top_history_list->destroy;
	}
}

# Called from destroy event mapping
sub histDestroy {
	my $widget = shift;
	if (defined($top_history_list) && (shift == $top_history_list)) {
		$top_history_list->overrideredirect(0);
		undef $top_history_list;
		$mw -> focus;
		$widget -> focus;
	}
}

#Center a toplevel on screen or above parent
sub centerList {
	my $w = shift;
	my $widget = shift;

	my $x = $widget -> x + $mw->rootx;
	my $y = $widget -> y + 20 + $mw->rooty;
	
	$x = int($x);
	$y = int($y);
	my $g = "+$x+$y";
	$w->geometry($g);	
}

############################## External Stuff #######################################
#Open the file in an external Editor
sub editor {
	my $file = shift;
	#File not available
	unless ($file) {
		error("No files specified");
		return 0;
	}

	if ($opt{'editor'}) {
		my $editor = $opt{'editor'};
		$editor =~ s/\//\\/g;

 		unless(system($editor,$file)) {
			error("Error: Cannot run command '$opt{'editor'} $file'.\n$?");
		}
	}
	else {
		error("No external editor defined. Set an Editor using 'Edit->Options->Editor'.");
	}
}
############################## Process Functions ####################################
#Returns true if the given element is in the given Array.
sub inArray {
	my $item = shift;
	my @array= shift;
	foreach my $element (@array) {
		return 1 if($element eq $item);
	}
	return 0;
}

#Return all the lines in the text area as an array of lines
sub getLines {
	my $data;
	if($gui_is_on) {
		$data = $stxt_text -> get('1.0','end');
		my @lines = split(/\n/,$data);
		return @lines;
	} else {
		return openFile();
	}
}
#Return the data in the text area
sub getData {
	return $stxt_text -> get('1.0','end');
}

#Show the line numbers of the file in the left margin
sub lineNumbersDisplay {
	$lst_line_numbers -> delete(0,'end');
	if(!$opt{'show_line_numbers'}) {
		$lst_line_numbers -> configure(-width=>0,-bg=>'#C0C0C0');
		return -1;
	}

	#Show Line numbers
	$lst_line_numbers -> configure(-width=>4,-bg=>'#808080');
	
	my @lines = getLines();

	$mw->update; #:DEBUG: - not actually needed.
	my $max_chars_per_line = int($stxt_text -> width/7) - 2;

	#Rearrange line numbers accoding to wrapping.
	my $line_number = 0;
	my $line_label = "";
	my $lines_to_skip;
	for my $line(@lines) {
		$line_number++;
		
		#How many spaces must be inserted before the number - this is to 
		#		align the line numbers to the right.
		# if($line_number<10) {
		# 	$space_multiplier=3
		# } elsif($line_number<100) {
		# 	$space_multiplier=2;
		# } elsif($line_number<1000) {
		# 	$space_multiplier=1;
		# } else {
		# 	$space_multiplier=0;
		# }
		# $line_label = " " x $space_multiplier . $line_number;

		$lst_line_numbers -> insert('end',$line_number);
		
		next if(!$opt{'word_wrap'});#If there is no word wrap, don't bother with the rest.
		#If a wrap has occured - put a small dot
		if(length($line) > $max_chars_per_line) {
			$lines_to_skip = int(length($line)/$max_chars_per_line);
			for(my $i=0;$i<$lines_to_skip;$i++) {
				$lst_line_numbers -> insert('end',".");
			}
		}
	}
}

#Inserts Escapes for all speical chars in the given reglar expression
sub escape {
	my $reg_exp = $ent_regexp -> get();
	
	if($reg_exp =~ /\\/) { 	#If the expression have a '\' char, do an escaped replace - replace the 
							# char only if the last char was not a '\'
		$reg_exp =~ s/([^\\])
		([
			\\\/\"\'
			\?\*\+\-\$\^
			\:\|\.\,
			\[\]\(\)\{\}
		])
		/\1\\\2/gx;
	} else {	#Just do the escaping - much safer
		$reg_exp =~ s/([
			\\\/\"\'
			\?\*\+\-\$\^
			\:\|\.\,
			\[\]\(\)\{\}
		])
		/\\\1/gx;
	}
	$ent_regexp -> delete(0,'end');
	$ent_regexp -> insert(0, $reg_exp);
}

#Filter the data based on the given info
sub filterData {
	my($from,$to) = split(/,/,$filter);
	my $data = getData();

	if($from =~ /^\d+$/) { #If it is made of just numbers
		$stxt_text -> delete('1.0',"$from.0");
	} elsif($from) { #It is a regexp
		if($data =~ /$from/g) { #If it matches
			#Find the position of the line that matchs the regexp
			my $position = pos $data;
			my @all_lines = split(/\n/,$data);
			
			my $current_line = 0;
			my $total_chars = 0;
			foreach my $l (@all_lines) {
				$total_chars += length($l);
				last if ($total_chars > $position);
				$current_line++;
			}
			
			#Now delete it
			$stxt_text -> delete('1.0',"$current_line.0");
		}
	}
	
	if($to =~ /^\d+$/) { #If it is made of just numbers
		$stxt_text -> delete("$to.0",'end');
	} elsif($to) {
		if($data =~ /$to/g) { #If it matches
			#Find the position of the line that matchs the regexp
			my $position = pos $data;
			my @all_lines = split(/\n/,$data);
			
			my $current_line = 0;
			my $total_chars = 0;
			foreach (@all_lines) {
				$total_chars += length;
				last if ($total_chars > $position);
				$current_line++;
			}
			$current_line++;
			#Now delete it
			$stxt_text -> delete("$current_line.0",'end');
		}
	}
}

#Process the file line by line. If a list is given in the argument, process all the regexps in the list.
sub sed {
	my @all_regexps = @_;
	return unless $file;
	
	if(@all_regexps){
	} elsif($regexp) {
		$all_regexps[0] = $regexp;

		#Save the regexp to the history list
		if(!inArray($regexp,@regexp_history)) {
			push(@regexp_history,$regexp);
		}
		if(!inArray($replace,@replace_history) and $replace) {
			push(@replace_history,$replace);
		}
	} else {
		return;#If a regular expression is not given, don't continue.
	}

	my @lines = getLines();
	my $data = "";
	my $this_line;
	my $direct_command = 0;
	
	if($opt{'direct_command'} == 1) {
		$direct_command = 1;
	} elsif($opt{'direct_command'} == 0) {
		$direct_command = 0;
	} else {
		#The direct command will be enabled if the regexp starts with 's/' or '/' and has another '/' in it.
		$direct_command = 1 if ($regexp =~ /^s?\/.+\//);
	}

	foreach my $reg (@all_regexps) {
		#Initializations
		$regexp = $reg;
		$data = "";
		my $line_count = $#lines;
		return unless $line_count;
		my $dot_points = int($line_count/10);

		for (my $i=0; $i<=$line_count; $i++) {
			$_ = $lines[$i];
	
			#Display the progress.
			if($dot_points and $gui_is_on) {
				if($i % $dot_points == 0 and $opt{'show_progress'}) {
					$percent_done += 10;
					$pbr_progress -> value($percent_done);
					$mw -> update; #Display the progress
				}
			}
		
			#If the direct_command mode is enabled, just run the command as they come from the 
			#	regexp field. This can be in the form of 's/<a href="(.+?)"/\1/gi;'
			#	The system will do the replace with just one field.
			if($direct_command) {
				if(eval("$regexp")) {
					next if($opt{'print_unmatched'});#Don't print if we are 'print_unmatched' flag is on
					$data .= $_ . "\n";
				} else {#If the the line did not match the regexp.
					if($opt{'print_all_lines'} or $opt{'print_unmatched'}) {
						$data .= $_ . "\n";
					}
				}
			}
			else {
				if(/$regexp/) {
					next if($opt{'print_unmatched'});#Don't print if we are 'print_unmatched' flag is on
					if($replace) { #If a replace text is there, do the replace.
						if($modifiers) {
							eval("s/$regexp/$replace/$modifiers;");#We have to use eval - otherwise the modifiers won't work.
						} else { #Non-eval version - too much javascript can do that to a man.
							s/$regexp/$replace/;
						}
					}
					$data .= $_ . "\n";
				} else { #If the the line did not match the regexp.
					if($opt{'print_all_lines'} or $opt{'print_unmatched'}) {
						$data .= $_ . "\n";
					}
				}
			}
		}

		@lines = split(/\n/,$data); #Save the data to the 'lines' variable - needed if more regexps must be done on it.
	}

	#Put the new lines in.
	if($gui_is_on) {
		$stxt_text -> delete('1.0','end');
		$stxt_text -> insert('1.0',$data);
	} else {
		return $data;
	}
}

############################## First And Last ##########################################
#Happens when the applications is closed
sub exiter {
	#Save the options to a file
	my $options_ok = open(OPT,">$opt{'option_file'}");
	if($options_ok) {
		# The option file is saved in the format of perl statments - like
		#	$opt{'show_line_numbers'} = "1";
		while (my($key,$value) = each %opt) {
    		print OPT "\$opt{'$key'}	= \"$value\";\n";
		}
	} else {
		error("Cannot save options to file '$opt{'option_file'}'...\n$!");
	}
	close(OPT);
	exit;
}

#Initialize the script
sub init {
	$but_filter -> focus;

	#Get Options
	# The option file is saved in the format of perl statments - like
	#	$opt{'show_line_numbers'} = "1";
	# Now all we have to do here is read the file and eval all lines in it.
	my $options_ok = open(OPT,$opt{'option_file'});
	if($options_ok) {
		eval while(<OPT>);
	}
	close(OPT);
	lineNumbersDisplay();
}

############################## To Do ##########################################
# Word Wrap On-Off		- Done
# Line Numbers On-Off	- Done
# Undo					- Done
# Temperory RegExp Storing ability	- Isn't History Enough
# Save as script		- Done
# Run script without gui- Done
# Save Options			- Done
# Previous replaces,regexps are saved - Done
# Execute a scirpt file after loading it - Done
